/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Plugin;
import org.eclipse.core.runtime.Preferences;
import org.eclipse.core.runtime.Status;

import com.example.autorefresh.internal.MonitorManager;
import com.example.autorefresh.internal.RefreshManager;

/**
 * The <code>AutoRefreshPlugin</code> is responsible for monitoring the
 * filesystem and synchronizing the workspace model with changes that
 * are detected.
 * <p>
 * <b>NOTE:</b> This plugin will not start automatically, it must be started by
 * calling #doStartup().
 * <p>
 * Clients who wish to override the default implementation of the filesystem
 * monitoring should register an <code>IMonitorFactory</code> extension.  See
 * <code>IMonitorFactory</code> for more details.
 * <p>
 * In the event that contributed <code>IMonitor</code> objects cannot monitor
 * a given volume, the auto-refresh support will fall back to a polling
 * mechanism to detect changes in the filesystem.  This mechanism can
 * be configured to not run by setting the auto.poll property to <code>false</code>.
 * When not in auto.poll mode, polling can still be requested by clients
 * by invoking the #poll() method. The polling mechanism also exposes the 
 * auto.poll.delay property which determines the amount of time 
 * (in milliseconds) between polls.
 * <p>
 * When differences are detected, the workspace model is synchronized with
 * the filesystem in the background.  This can be configured to not occur
 * by setting the auto.refresh property to <code>false</code>.  When auto.refresh
 * has been set to <code>false</code>, changes are accumulated until the
 * #refresh(IProgressMonitor) method is invoked.
 * <p>
 * <b>NOTE:</b>  Background refreshing locks the workspace tree causing other 
 * operations to block.  During this time changes can be made to the 
 * contents of buffers, but saves, builds, etc. will block.
 * <p>
 * Known limititations:
 * <ul>
 * <li>Potential blocking during background refresh, see above.</li>
 * <li>Deletion of projects</li>
 * </ul>
 * <p>
 * WARNING:  This is experimental API.  It is likely to change in future releases.
 */
public class AutoRefreshPlugin extends Plugin {
	/**
	 * The singleton instance of the <code>AutoRefreshPlugin</code>.
	 */
	private static AutoRefreshPlugin fDefault;
	
	/**
	 * Property identifier for the auto polling flag
	 */
	public static final String PREF_AUTO_POLL= "auto.poll";

	/**
	 * Property identifier for the auto refreshing flag
	 */
	public static final String PREF_AUTO_REFRESH= "auto.refresh";
	/**
	 * Property identifier for the refresh delay field
	 */
	public static final String PREF_AUTO_REFRESH_DELAY= "auto.refresh.delay";
	/**
	 * Hardcoded default for refresh delay, set to 10 seconds.
	 */
	public static final long DEFAULT_REFRESH_DELAY= 10 * 1000;
	/**
	 * Property identifier for the refresh delay field
	 */
	public static final String PREF_AUTO_POLL_DELAY= "auto.poll.delay";
	/**
	 * Hardcoded default for refresh delay, set to 10 seconds.
	 */
	public static final long DEFAULT_POLL_DELAY= 10 * 1000;
	/**
	 * The error code indicating an internal exception has occured
	 */
	private static final int INTERNAL_EXCEPTION= 1;
	/**
	 * Extension point name for the monitorFactories extension point.
	 */
	private static final String EXTENSION_POINT_MONITOR_FACTORIES= "monitorFactories";
	/**
	 * Extension attribute name for specifying which class to instantiate
	 * to create a monitor factory.
	 */
	private static final String CLASS= "class";
	
	/**
	 * The list of registered monitor factories.
	 */
	private IMonitorFactory[] fMonitorFactories;
	/**
	 * The monitor manager is responsible for doing periodic
	 * (or on demand) filesystem monitoring.
	 */
	private MonitorManager fMonitorManager;
	/**
	 * The refresh manager is responsible for doing periodic
	 * (or on demand) refreshes.
	 */
	private RefreshManager fRefreshManager;
	/*
	 * @see org.eclipse.core.runtime.Plugin#Plugin(IPluginDescriptor)
	 */
	public AutoRefreshPlugin(IPluginDescriptor descriptor) {
		super(descriptor);
		fDefault= this;
		//setDebugging(true);
	}
		
	/**
	 * Performs startup code.  This method must be called after
	 * plugin startup to avoid potential createExecutableExtension
	 * failures.
	 * <p>
	 * <b>NOTE:</b> This method intentionally does not override Plugin#startup
	 * in order to avoid problems with plugin startup and createExecutableExtension.
	 */
	public void doStartup() {
		fMonitorManager= new MonitorManager();
		fMonitorManager.startup();
		fRefreshManager= new RefreshManager();
		fRefreshManager.start();
	}

	/**
	 * Answers the singleton instance of the <code>AutoRefreshPlugin</code>.
	 */
	public static AutoRefreshPlugin getDefault() {
		return fDefault;
	}
	
	/*
	 * @see org.eclipse.core.runtime.Plugin#initializeDefaultPluginPreferences()
	 */
	protected void initializeDefaultPluginPreferences() {
		Preferences preferences= getPluginPreferences();
		preferences.setDefault(PREF_AUTO_REFRESH, true);
		preferences.setDefault(PREF_AUTO_POLL, true);
		preferences.setDefault(PREF_AUTO_REFRESH_DELAY, DEFAULT_REFRESH_DELAY);
		preferences.setDefault(PREF_AUTO_POLL_DELAY, DEFAULT_POLL_DELAY);
	}	

	/**
	 * Logs the exception as an internal exception.
	 * 
	 * @param throwable	A throwable that was caught, can be <code>null</code>.
	 * @param message		A message describing the exception, can be <code>null</code>.
	 */
	public static void log(Throwable throwable, String message) {
		fDefault.getLog().log(new Status(
			Status.ERROR,
			fDefault.getDescriptor().getUniqueIdentifier(), 
			INTERNAL_EXCEPTION, 
			message,
			throwable));
		if (getDefault().isDebugging()) {
			System.err.println("An error occured:");
			if (message != null) {
				System.err.println(message);
			}
			if (throwable != null) {
				throwable.printStackTrace();
			}
		}
	}

	/**
	 * Requests that the resources in the request be refreshed.  Does
	 * nothing if the request is <code>null</code>.  If PREF_AUTO_REFRESH is
	 * <code>false</code> stores the request for a later call to #refresh
	 * (IProgressMonitor).
	 * 
	 * @param request	A refresh request, can be <code>null</code>.
	 * @param monitor	A progress monitor, can be <code>null</code>.
	 */
	public void refresh(IRefreshRequest request, IProgressMonitor monitor) {
		if (fRefreshManager != null) {
			fRefreshManager.addRequest(request);
		}
	}
	
	/**
	 * Sets the refresh delay field.  The refresh delay is the 
	 * number of milliseconds the <code>AutoRefreshPlugin</code>
	 * waits before updating the workspace with the changes that have
	 * occured in the filesystem.
	 * @param delay	The number of milliseconds to delay between
	 * 	refreshes.  If this number is less than or equal to zero, the 
	 * delay will be set to the default value.
	 */
	public void setRefreshDelay(long delay) {
		Preferences preferences= getPluginPreferences();
		if (delay <= 0) {
			preferences.setToDefault(PREF_AUTO_REFRESH_DELAY);
		} else {
			long delayPreference= getPluginPreferences().getLong(PREF_AUTO_REFRESH_DELAY);
			if (delayPreference != delay) {
				preferences.setValue(PREF_AUTO_REFRESH_DELAY, delay);
			} else {
				return;
			}
		}
		savePluginPreferences();		
	}


	/*
	 * @see org.eclipse.core.runtime.Plugin#shutdown()
	 */
	public void shutdown() throws CoreException {
		fMonitorManager.shutdown();
		fMonitorManager= null;
		
		fRefreshManager.stop();
		fRefreshManager= null;
		
		if (getPluginPreferences().needsSaving()) {
			savePluginPreferences();
		}
	}

	/**
	 * Queries the extensions of the monitorFactories extension point, and
	 * creates the factory classes. Will never return <code>null</code>.
	 * 
	 * @return IMonitorFactory[] The array of registered <code>IMonitorFactory</code> objects or an empty array.
	 */
	public IMonitorFactory[] getMonitorFactories() {
		if (fMonitorFactories != null)
			return fMonitorFactories;
		IPluginDescriptor descriptor= AutoRefreshPlugin.getDefault().getDescriptor();
		IExtensionPoint extensionPoint= descriptor.getExtensionPoint(EXTENSION_POINT_MONITOR_FACTORIES);
		IConfigurationElement[] infos= extensionPoint.getConfigurationElements();
		List factories= new ArrayList(infos.length);
		for (int i= 0; i < infos.length; i++) {
			IConfigurationElement configurationElement= infos[i];
			IMonitorFactory factory= null;
			try {
				factory= (IMonitorFactory)configurationElement.createExecutableExtension(CLASS);
			} catch (CoreException e) {
				log(e, "While trying to create an IMonitorFactory for " + configurationElement.getAttribute(CLASS));
			}
			if (factory != null) {
				factories.add(factory);
			}
		}
		if (!factories.isEmpty()) {
			fMonitorFactories= new IMonitorFactory[factories.size()];
			factories.toArray(fMonitorFactories);
		} else {
			fMonitorFactories= new IMonitorFactory[0];
		}
		return fMonitorFactories;
	}
	
	/**
	 * Requests that a polling of the filesystem take place.
	 */
	public void poll() {
		if (fMonitorManager != null) {
			fMonitorManager.poll();
		}
	}
	
	/**
	 * Requests that all stored refresh requests get processed.
	 */
	public void refresh(IProgressMonitor monitor) {
		if (fRefreshManager != null) { 
			fRefreshManager.refresh(monitor);
		}
	}
	
	/**
	 * Notifies the <code>AutoRefreshPlugin</code> that the given
	 * monitor has encountered a failure from which it cannot recover
	 * while monitoring the given container.
	 * <p>
	 * If the given container is <code>null</code> the <code>AutoRefreshPlugin</code>
	 * will assume that the monitor has failed completely, and take over
	 * the monitoring responsibilities for all containers that the
	 * monitor was monitoring.
	 * 
	 * @param monitor		A monitor which has encountered a failure
	 * 						that it cannot recover from, can not be
	 * 						<code>null</code>.
	 * @param container	The container that the monitor can no longer
	 * 						monitor, or <code>null</code> to indicate that
	 * 						the monitor can no longer monitor any of the
	 * 						containers it was monitoring.
	 */
	public void monitorFailed(IMonitor monitor, IContainer container) {
		if (isDebugging()) {
			System.err.println(monitor + " failed for container " + container);
		}
		if (fMonitorManager != null) {
			fMonitorManager.monitorFailed(monitor, container);
		}
	}
}
