/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;

/**
 * The <code>SimpleRefreshRequest</code> is an implementation
 * of the <code>IRefreshRequest</code> interface used by clients
 * to make refresh requests on single resources which are managed
 * locally.  
 * 
 * Clients should implement <code>IRefreshRequest</code> for themselves 
 * if they have resources which are managed remotely or have 
 * <code>IMonitor</code> objects which can return multiple resources
 * in a single callback.
 * <p>
 * WARNING:  This is experimental API.  It is likely to change in future releases.
 * 
 * @see com.example.autorefresh.IRefreshRequest
 */
public abstract class SimpleRefreshRequest implements IRefreshRequest {
	/**
	 * The resources to be refreshed.
	 */
	protected IResource[] fResources;
	
	/**
	 * Request type constant for file additions.
	 */
	public static final int ADD = 1;
	/**
	 * Request type constant for file deletions.
	 */
	public static final int DELETE = 2;
	/**
	 * Request type constant for file changes.
	 */
	public static final int CHANGE = 3;
	/**
	 * Request type constant for file transforms, for example, a file
	 * becomes a directory.
	 */
	public static final int TRANSFORM= 4;
	
	/**
	 * Request type constant for requests to refresh
	 * a single file.
	 */
	public static final int DEPTH_ZERO= CHANGE;
	/**
	 * Request type constant for request to refresh
	 * a single file and its direct children.
	 */
	public static final int DEPTH_ONE= ADD;
	/**
	 * Request type constant for request to refresh
	 * the entire subtree starting with the given file.
	 */
	public static final int DEPTH_INFINITE= TRANSFORM;
	/**
	 * An empty resource array.
	 */
	private static IResource[] EMPTY_RESOURCE_ARRAY= new IResource[0];
	
	/**
	 * Creates a new <code>SimpleRefreshRequest</code>
	 * @param resources the resources to be refreshed
	 */
	private SimpleRefreshRequest(IResource[] resources) {
		fResources= resources;
	}
	
	/**
	 * Creates a new <code>SimpleRefreshRequest</code> based off of the path
	 * to the resource(s) and the request type.
	 * 
	 * @param absolutePath		An absolute path to a local file, can be
	 * <code>null</code>.
	 * @param requestType		One of ADD, DELETE, CHANGE, TRANSFORM,
	 * DEPTH_ZERO, DEPTH_ONE, DEPTH_INFINTE.
	 * @return SimpleRefreshRequest	A new request, or <code>null</code> if the
	 * resource(s) for the request could not be found.
	 */
	public static SimpleRefreshRequest createRequest(String absolutePath, int requestType) {
		if (absolutePath == null || absolutePath.length() == 0) {
			return null;
		}
		IResource[] resources= null;
		String filename= null;
		if (requestType == ADD) {
			IPath fullPath= new Path(absolutePath);
			filename= fullPath.lastSegment(); 
			resources= findResource(fullPath.removeLastSegments(1));
		} else {
			resources= findResource(new Path(absolutePath));
		}
		if (resources.length == 0) {
			return null;
		}
		switch (requestType) {
			case ADD:
				return new AddRefreshRequest(resources, filename);
			case CHANGE:
				return new ChangeRefreshRequest(resources);
			case DELETE:
				return new DeleteRefreshRequest(resources);
			case TRANSFORM:
				return new TransformRefreshRequest(resources);
		}
		return null;
	}
	
	/*
	 * Finds the resources that map to the given path.  Returns
	 * an empty array of IResource objects if no resources exist
	 * which map to the given path.
	 */
	private static IResource[] findResource(IPath path) {
		IWorkspace workspace= ResourcesPlugin.getWorkspace();
		if (workspace != null) {
			IWorkspaceRoot root= workspace.getRoot();
			if (root != null) {
				IContainer[] containers= root.findContainersForLocation(path);
				if (containers.length == 0) {
					IFile[] files= root.findFilesForLocation(path);
					if (files.length == 0) {
						return EMPTY_RESOURCE_ARRAY;
					}
					return files;
				}
				return containers;
			}
		}
		return null;
	}	
	
	/**
	 * A refresh request which represents the addition of a single
	 * file or folder.
	 */
	private static class AddRefreshRequest extends SimpleRefreshRequest {
		/*
		 * The simple filename of the file or folder that was added.
		 */
		private String fFilename;
		
		/**
		 * Creates an <code>AddRefreshRequest</code>.
		 * @param parent	The parent folder which contains the new 
		 * 					file or folder, can not be <code>null</code>.
		 * @param filename	The simple filename of the new file or folder,
		 * 					can not be <code>null</code>.
		 */
		public AddRefreshRequest(IResource[] parents, String filename) {
			super(parents);
			fFilename= filename;
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#hasAdditions()
		 */
		public boolean hasAdditions() {
			return true;
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#getAdditions()
		 */
		public Map getAdditions() {
			Map map= new HashMap(1);
			List list= new ArrayList(1);
			for (int i = 0; i < fResources.length; i++) {
				map.put(fResources[i], list);
			}
			list.add(fFilename);
			return map;
		}
	}
	
	/**
	 * A refresh request which represents the deletion of a single
	 * file or folder.
	 */	
	private static class DeleteRefreshRequest extends SimpleRefreshRequest {
		/**
		 * Creates a <code>DeleteRefreshRequest</code>.
		 * @param resources	The resources that were deleted, can not be
		 * <code>null</code>.
		 */
		public DeleteRefreshRequest(IResource[] resources) {
			super(resources);
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#hasDeletions()
		 */
		public boolean hasDeletions() {
			return true;
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#getDeletions()
		 */
		public IResource[] getDeletions() {
			return fResources;
		}
	}
	
	/**
	 * A refresh request which represents the change of a single file
	 * or folder.
	 */
	private static class ChangeRefreshRequest extends SimpleRefreshRequest {
		/**
		 * Creates a <code>ChangeRefreshRequest</code>.
		 * @param resources	The resources which were changed, can not be
		 * <code>null</code>.
		 */
		public ChangeRefreshRequest(IResource[] resources) {
			super(resources);
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#hasChanges()
		 */
		public boolean hasChanges() {
			return true;
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#getChanges()
		 */
		public IResource[] getChanges() {
			return fResources;
		}
	}
	
	/**
	 * A refresh request which represents the transform of a single file
	 * or folder.
	 */	
	private static class TransformRefreshRequest extends SimpleRefreshRequest {
		/**
		 * Creates a <code>TransformRefreshRequest</code>.
		 * @param resources	The resources which were transformed, can not be
		 * <code>null</code>.
		 */
		public TransformRefreshRequest(IResource[] resources) {
			super(resources);
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#hasTransforms()
		 */
		public boolean hasTransforms() {
			return true;
		}
		
		/*
		 * @see com.example.autorefresh.IRefreshRequest#getTransforms()
		 */
		public IResource[] getTransforms() {
			return fResources;
		}
	}		
	
	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasAdditions()
	 */
	public boolean hasAdditions() {
		return false;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getAdditions()
	 */
	public Map getAdditions() {
		return null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasDeletions()
	 */
	public boolean hasDeletions() {
		return false;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getDeletions()
	 */
	public IResource[] getDeletions() {
		return null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasChanges()
	 */
	public boolean hasChanges() {
		return false;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getChanges()
	 */
	public IResource[] getChanges() {
		return null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasTransforms()
	 */
	public boolean hasTransforms() {
		return false;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getTransforms()
	 */
	public IResource[] getTransforms() {
		return null;
	}
	
	/*
	 * @see com.example.autorefresh.IRefreshRequest#isEmpty()
	 */
	public boolean isEmpty() {
		return false;
	}	
}
