/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.internal;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IPathVariableChangeEvent;
import org.eclipse.core.resources.IPathVariableChangeListener;
import org.eclipse.core.resources.IPathVariableManager;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceChangeEvent;
import org.eclipse.core.resources.IResourceChangeListener;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IResourceProxy;
import org.eclipse.core.resources.IResourceProxyVisitor;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;

import com.example.autorefresh.AutoRefreshPlugin;
import com.example.autorefresh.IMonitor;
import com.example.autorefresh.IMonitorFactory;

/**
 * Manages monitors by creating new monitors when projects are added
 * and removing monitors when projects are removed.  Also handles
 * the polling mechanism when contributed native monitors cannot handle
 * a project.
 */
public class MonitorManager {
	/**
	 * A mapping of IResource to collection of IMonitor.
	 */
	private Map fRegisteredMonitors;
	/**
	 * A <code>PollingManager</code> for handling filesystem polling.
	 */
	private PollingManager fPollingManager;
	/**
	 * A resource change listener which handles adding and removing monitors
	 * when resources are added/removed/opened/closed.
	 */
	private IResourceChangeListener fResourceListener;
	/**
	 * A path variable change listener which handles adding and removing
	 * monitors when path variables change.
	 */
	private IPathVariableChangeListener fPathVariableListener;
	
	public MonitorManager() {
		fRegisteredMonitors= Collections.synchronizedMap(new HashMap(10));
		fPollingManager= new PollingManager();
	}

	private void monitor(IResource resource) {
		if (resource == null) {
			return;
		}
		if (isMonitoring(resource)) {
			return;
		}
		IPath path= resource.getLocation();
		if (path == null) {
			return;
		}
		boolean pollingMonitorNeeded= true;
		IMonitorFactory[] factories= AutoRefreshPlugin.getDefault().getMonitorFactories();
		if (factories != null) {
			for (int i= 0; i < factories.length; i++) {
				IMonitor monitor= factories[i].createMonitor(resource);
				if (monitor != null) {
					registerMonitor(monitor, resource);
					pollingMonitorNeeded= false;
				}	
			}
		}
		if (pollingMonitorNeeded) {
			addPollingMonitor(new PollingMonitor(resource), resource);
		}
	}
	
	private void unmonitor(IResource resource) {
		if (resource == null)
			return;
		if (!isMonitoring(resource))
			return;
		fireUnmonitor(resource);
		synchronized (fRegisteredMonitors) {
			fRegisteredMonitors.remove(resource);
		}
	}

	public void startup() {
		addResourceListener();
		addPathVariableListener();
		for (Iterator i = getResourcesToMonitor().iterator(); i.hasNext();) {
			monitor((IResource) i.next());
		}
		
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Startup Monitor Manager");
		}		
	}


	/**
	 * Registers a <code>IPathVariableChangeListener</code>.  This listener is
	 * responsible for managing monitors on paths that contain variables.
	 */
	private void addPathVariableListener() {
		fPathVariableListener = new IPathVariableChangeListener() {
			/**
			 * @see org.eclipse.core.resources.IPathVariableChangeListener#pathVariableChanged(org.eclipse.core.resources.IPathVariableChangeEvent)
			 */
			public void pathVariableChanged(IPathVariableChangeEvent event) {
			}
		};
		IWorkspace workspace= ResourcesPlugin.getWorkspace();
		if (workspace != null) {
			IPathVariableManager manager= workspace.getPathVariableManager();
			if (manager != null) {
				manager.addChangeListener(fPathVariableListener);
			}
		}
	}

	private List getResourcesToMonitor() {
		final List resourcesToMonitor= new ArrayList(10);
		IWorkspace workspace= ResourcesPlugin.getWorkspace();
		if (workspace == null) {
			return resourcesToMonitor;
		}
		
		IWorkspaceRoot root= workspace.getRoot();
		if (root == null) {
			return resourcesToMonitor;
		}
		
		IProject[] projects= root.getProjects();
		if (projects == null || projects.length == 0) {
			return resourcesToMonitor;
		}
		
		IResourceProxyVisitor visitor= new IResourceProxyVisitor() {
			/**
			 * A fast visitor that looks at just one level of resources for
			 * linked resources.  Currently, linked resources can only be
			 * direct children of projects.
			 * 
			 * @see org.eclipse.core.resources.IResourceProxyVisitor#visit(org.eclipse.core.resources.IResourceProxy)
			 */
			public boolean visit(IResourceProxy proxy)
				throws CoreException {
				if (proxy.getType() == IResource.PROJECT) {
					return true;
				}
				if (proxy.isLinked()) {
					resourcesToMonitor.add(proxy.requestResource());
				}
				return false;
			}
		};
		for (int i= 0; i < projects.length; i++) {
			IProject project= projects[i];
			resourcesToMonitor.add(project);
			try {
				project.accept((IResourceProxyVisitor)visitor, IResource.NONE);
			} catch (CoreException e) {
				AutoRefreshPlugin.log(e, "Could not visit project");
			}
		}
		
		return resourcesToMonitor;
	}
	
	public void shutdown() {
		fPollingManager.shutdown();
		removeResourceListener();
		removePathVariableListener();
		synchronized (fRegisteredMonitors) {
			for (Iterator i = fRegisteredMonitors.keySet().iterator(); i.hasNext();) {
				IContainer container= (IContainer) i.next();
				fireUnmonitor(container);
			}
		}
		synchronized (this) {
			fRegisteredMonitors= null;
		}
		fPollingManager= null;
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Shutdown Refresh Manager");
		}		
	}
	
	/**
	 * Removes the path variable listener.
	 */
	private void removePathVariableListener() {
		if (fPathVariableListener != null) {
			IWorkspace workspace= ResourcesPlugin.getWorkspace();
			if (workspace != null) {
				IPathVariableManager manager= workspace.getPathVariableManager();
				if (manager != null) {
					manager.removeChangeListener(fPathVariableListener);
				}
			}
			fPathVariableListener= null;
		}
	}

	private void registerMonitor(IMonitor monitor, IResource resource) {
		List list= (List)fRegisteredMonitors.get(resource);
		if (list == null) {
			list= new ArrayList(1);
			synchronized (fRegisteredMonitors) {
				fRegisteredMonitors.put(resource, list);
			}
		}
		list.add(monitor);
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Added monitor (" + monitor + ") for " + resource);
		}					
	}

	private void addPollingMonitor(PollingMonitor monitor, IResource resource) {
		registerMonitor(monitor, resource);
		monitor.setManager(fPollingManager);
		fPollingManager.start();
	}
	
	private void removeMonitor(IMonitor monitor, IContainer container) {
		List list= (List)fRegisteredMonitors.get(container);
		if (list != null && !list.isEmpty()) {
			synchronized (fRegisteredMonitors) {
				list.remove(monitor);
			}
		}
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Removed monitor (" + monitor + ") for " + container);
		}					
	}
	
	private void fireUnmonitor(IResource resource) {
		List monitors= (List)fRegisteredMonitors.get(resource);
		if (monitors != null) {
			synchronized (fRegisteredMonitors) {
				if (!monitors.isEmpty()) {			
					for (Iterator i= monitors.iterator(); i.hasNext();) {
						IMonitor monitor= (IMonitor) i.next();
						monitor.unmonitor(resource);
					}
				}
			}
		}
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Removed monitors for " + resource);
		}					
	}		
	
	/**
	 * Adds a resource listener that listens for project
	 * adds/removes and notifies the <code>MonitorManager</code>.
	 * <p>
	 * Tracks opened/closed state of projects like they were
	 * added/removed.
	 * <p>
	 * Also listens for linked folders which are added and removed
	 */
	private void addResourceListener() {
		fResourceListener= new IResourceChangeListener() {
			IResourceDeltaVisitor fVisitor= new IResourceDeltaVisitor() {
				public boolean visit(IResourceDelta delta) throws CoreException {
					IResource resource= delta.getResource();
					switch (resource.getType()) {
						case IResource.FILE:
						case IResource.FOLDER:
							if (resource.isLinked()) {
								switch (delta.getKind()) {
									case IResourceDelta.ADDED:
										monitor(resource);
										break;
									case IResourceDelta.REMOVED:
										unmonitor(resource);
										break;
									default:
										break;
								}
							}
							return false;
						case IResource.ROOT:
							return true;		
						case IResource.PROJECT:
							IProject project= (IProject)resource;
							switch (delta.getKind()) {
								case IResourceDelta.ADDED:
									/*
									 * Project deletion is handled in resourceChanged(IResourceEvent)
									 */
									if (project.isOpen()) {
										monitor(project);
									}
									break;
								case IResourceDelta.CHANGED:
									if ((delta.getFlags() & IResourceDelta.OPEN) != 0) {
										/*
										 * Project closure is handled in resourceChanged(IResourceEvent)
										 */
										if (project.isOpen()) {
											monitor(project);
										}
									}
									break;
								default:
									break;
							}
							return true;
						default:
							break;
					}
					return false;
				}
			};
			public void resourceChanged(IResourceChangeEvent event) {
				switch (event.getType()) {
					case IResourceChangeEvent.PRE_DELETE:
					case IResourceChangeEvent.PRE_CLOSE:
						/*
						 * Additions and project open handled in visitor.
						 */
						unmonitor((IProject)event.getResource());
						break;
					default:
						try {
							event.getDelta().accept(fVisitor);					
						} catch (CoreException e) {
							AutoRefreshPlugin.log(e, "During resource callback");
						}
				}
			}
		};
		IWorkspace workspace= ResourcesPlugin.getWorkspace();
		if (workspace != null) {
			workspace.addResourceChangeListener(fResourceListener, IResourceChangeEvent.PRE_DELETE | IResourceChangeEvent.POST_CHANGE);
		}
	}

	private void removeResourceListener() {
		if (fResourceListener != null) {
			IWorkspace workspace= ResourcesPlugin.getWorkspace();
			if (workspace != null) {
				workspace.removeResourceChangeListener(fResourceListener);
			}
			fResourceListener= null;
		}
	}
	
	private boolean isMonitoring(IResource resource) {
		List list= (List)fRegisteredMonitors.get(resource);
		if (list == null)
			return false;
		return !list.isEmpty();
	}
	
	/**
	 * Notifies the manager that the given monitor has encountered a 
	 * failure from which it cannot recover while monitoring the 
	 * given container.
	 * <p>
	 * If the given container is <code>null</code> the manager
	 * will assume that the monitor has failed completely, and take over
	 * the monitoring responsibilities for all containers that the
	 * monitor was monitoring.
	 * 
	 * @param monitor		A monitor which has encountered a failure
	 * 						that it cannot recover from, can not be
	 * 						<code>null</code>.
	 * @param container	The container that the monitor can no longer
	 * 						monitor, or <code>null</code> to indicate that
	 * 						the monitor can no longer monitor any of the
	 * 						containers it was monitoring.
	 */
	public void monitorFailed(IMonitor monitor, IContainer container) {
		if (fRegisteredMonitors == null) {
			return;
		}
		PollingMonitor pollingMonitor= null;
		if (container == null) {
			Set containers= fRegisteredMonitors.keySet();
			synchronized (this) {
				if (fRegisteredMonitors == null) {
					return;
				}
				synchronized (fRegisteredMonitors) {	
					for (Iterator i = containers.iterator(); i.hasNext();) {
						container= (IContainer) i.next();
						List list= (List)fRegisteredMonitors.get(container);
						if (list != null && !list.isEmpty()) {
							if (list.contains(monitor)) {
								removeMonitor(monitor, container);
								pollingMonitor= new PollingMonitor(container); 
								addPollingMonitor(pollingMonitor, container);
							}
						}
					}
				}
			}
		} else {
			removeMonitor(monitor, container);
			addPollingMonitor(new PollingMonitor(container), container);
		}
	}

	public void poll() {
		fPollingManager.poll();
	}		
}
