/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.internal;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.Preferences;
import org.eclipse.core.runtime.Preferences.PropertyChangeEvent;

import com.example.autorefresh.AutoRefreshPlugin;

/**
 * The <code>PollingManager</code> is responsible for polling
 * the filesystem periodically.
 */
public class PollingManager implements Runnable {
	private List fPollingMonitors;
	private List fAddRemoveRequests;
	private boolean fRunning;
	
	/*
	 * The property change listener
	 */
	private Preferences.IPropertyChangeListener fPropertyChangeListener;
	
	public PollingManager() {
		fPollingMonitors= new ArrayList(1);
		fAddRemoveRequests= new ArrayList(1);
		fRunning= false;
		
		fPropertyChangeListener= new Preferences.IPropertyChangeListener() {
			/**
			 * Notifies the <code>AutoRefreshPlugin</code> when the PREF_AUTO_POLL
			 * property changes.
			 *
			 * @see IPropertyChangeListener#propertyChange(PropertyChangeEvent)
			 */
			public void propertyChange(PropertyChangeEvent event) {
				if (event.getProperty().equals(AutoRefreshPlugin.PREF_AUTO_POLL)) {
					manageAutoPolling();
				}
			}
		};
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		preferences.addPropertyChangeListener(fPropertyChangeListener);
	}
	
	private void manageAutoPolling() {
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		if (preferences.getBoolean(AutoRefreshPlugin.PREF_AUTO_POLL)) {
			start();
		} else {
			stop();
		}
	}
			
	public void addPollRequest(PollingMonitor monitor) {
		if (fRunning) {
			synchronized (fAddRemoveRequests) {
				fAddRemoveRequests.add(new AddRemoveRequest(monitor, true));
			}
		} else {
			fPollingMonitors.add(monitor);
		}
	}

	public void removePollRequest(PollingMonitor monitor) {
		if (fRunning) {
			synchronized (fAddRemoveRequests) {
				fAddRemoveRequests.add(new AddRemoveRequest(monitor, false));
			}
		} else {
			fPollingMonitors.remove(monitor);
		}
	}

	public void poll() {
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Polling: " + this);
		}
		synchronized (fPollingMonitors) {
			// creates minimal garbage in the common case
			// first see if we have any changes to the list of monitors
			List requests= null;
			synchronized (fAddRemoveRequests) {
				if (!fAddRemoveRequests.isEmpty()) {
					requests= fAddRemoveRequests;
					fAddRemoveRequests= new ArrayList(1);
				}
			}
			if (requests != null) {
				// process adds & removes in order
				for (int i= 0, length= requests.size(); i < length; i++) {
					AddRemoveRequest request = (AddRemoveRequest) requests.get(i);
					if (request.isAdd()) {
						fPollingMonitors.add(request.getMonitor());
					} else {
						fPollingMonitors.remove(request.getMonitor());
					}
				}
			}
			// do polling here
			for (int i= 0, length= fPollingMonitors.size(); i < length; i++) {
				PollingMonitor monitor= (PollingMonitor)fPollingMonitors.get(i);
				monitor.poll();
			}
		}
	}
	
	public void run() {
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		if (preferences.getBoolean(AutoRefreshPlugin.PREF_AUTO_POLL)) {
			while(fRunning && !fPollingMonitors.isEmpty()) {
				long delay= AutoRefreshPlugin.getDefault().getPluginPreferences().getLong(AutoRefreshPlugin.PREF_AUTO_POLL_DELAY);
				try {
					synchronized(this) {
						wait(delay);
					}
				} catch (InterruptedException e) {
					break;
				}
				if (fRunning && !fPollingMonitors.isEmpty()) {
					poll();
				}
			}
		}
		fRunning= false;
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Polling thread exiting");
		}		
	}
	
	public void start() {
		if (!fRunning) {
			fRunning= true;
			Thread t= new Thread(this, "Filesystem polling thread");
			int priority= t.getPriority() - 1;
			priority= (priority < Thread.MIN_PRIORITY) ? Thread.MIN_PRIORITY : priority; 
			t.setPriority(priority);
			t.start();
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Started polling thread.");
			}
		}
	}
	
	public void stop() {
		if (fRunning) {
			fRunning= false;
			synchronized (this) {
				notifyAll();
			}
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Stopped polling thread.");
			}
		}		
	}
	
	public void shutdown() {
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		preferences.removePropertyChangeListener(fPropertyChangeListener);
		fPropertyChangeListener= null;
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Shutdown polling manager.");
		}
	}
							
	/**
	 * Method isRunning.
	 * @return boolean
	 */
	public boolean isRunning() {
		return fRunning;
	}
	
	
	/**
	 * A simple data class for storing adds and removes.
	 */
	private static class AddRemoveRequest {
		private PollingMonitor fMonitor;
		private boolean fIsAdd;

		public AddRemoveRequest(PollingMonitor monitor, boolean isAdd) {
			fMonitor= monitor;
			fIsAdd= isAdd;
		}

		public PollingMonitor getMonitor() {
			return fMonitor;
		}

		public boolean isAdd() {
			return fIsAdd;
		}
	}	
}
