/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.internal;

import java.io.File;

import org.eclipse.core.internal.resources.ResourceInfo;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;

import com.example.autorefresh.AutoRefreshPlugin;
import com.example.autorefresh.IMonitor;

/**
 * The <code>PollingMonitor</code> is an <code>IMonitor</code> that
 * polls the filesystem rather than registering natively for callbacks.
 */
public class PollingMonitor implements IMonitor {
	private boolean fRunning;
	private PollingManager fManager;
	private IResource fRoot;
	
	/**
	 * Creates a new polling monitor.
	 * @param root		The root container to be monitored, not <code>null</code>.
	 * @param manager	The monitor manager, not <code>null</code>.
	 */
	public PollingMonitor(IResource resource) {
		fRoot= resource;
	}
	
	public void setManager(PollingManager manager) {
		fManager= manager;
		fManager.addPollRequest(this);		
	}
	
	/**
	 * Polls the filesystem under the root container for changes.
	 */
	public void poll() {
		long time= 0;
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Polling " + fRoot);
			time= System.currentTimeMillis();
		}
		fRunning= true;
		PollingResult result= null;
		IResource resource= fRoot;
		if (resource == null && AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Could not find resource for " + fRoot);
		}
		if (resource != null) {
			IPath path= resource.getLocation();
			if (path != null) {
				result= new PollingResult();
				checkResource(resource, path.toFile(), result);
			}
		} else {
			AutoRefreshPlugin.log(null, "Could not find " + fRoot);
		}
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			time= System.currentTimeMillis() - time;
			System.out.println("Poll of " + fRoot + " complete in " + time + "ms.");
		}		
		if (result != null && !result.isEmpty()) {
			AutoRefreshPlugin.getDefault().refresh(result, null);
		}
	}
	
	/*
	 * @see com.example.autorefresh.IMonitor#unmonitor(IContainer)
	 */
	public void unmonitor(IResource container) {
		if (container == null || container.equals(fRoot)) {
			fRunning= false;
			fManager.removePollRequest(this);
		}
	}


	/**
	 * @return boolean	<code>true</code>, if this monitor is polling.
	 */
	private boolean isRunning() {
		return fRunning;
	}

	/**
	 * Checks the <code>resource</code> against the <code>file</code>.
	 */
	private void checkResource(IResource resource, File file, PollingResult result) {
		if (resource instanceof IContainer) {
			checkContainer((IContainer)resource, file, result);
		} else {
			if (resource instanceof IFile) {
				checkFile((IFile)resource, file, result);
			}
		}
	}

	/*
	 * Checks to see if an members of the container have been added, changed, or
	 * deleted or if the container itself has transformed.
	 * 
	 * @param container		A container, not <code>null</code>.
	 * @param file			The file that corresponds to the container, 
	 * 						not <code>null</code>.
	 * @param result		A result to record changes, not <code>null</code>.
	 */
	private void checkContainer(IContainer container, File file, PollingResult result) {
		if (!file.isDirectory()) {
			result.recordTransform(container);
			return;
		}

		IResource[] members= null;
		try {
			members= container.members();
		} catch (CoreException e) {
			checkContainerExistence(container, file, result);
			return;
		}
		
		String[] list= file.list();
		if (list != null) {
			int hits= 0;
			for (int i = 0; isRunning() && i < members.length; i++) {
				IResource member= members[i];
				if (!searchInResource(member, file, list, result)) {
					result.recordDeletion(member);
				} else {
					hits++;
				}
			}
		
			if (isRunning() && hits != list.length) {
				for (int i = 0; i < list.length; i++) {				
					if (!resourceExists(list[i], members)) {
						result.recordAddition(container, list[i]);
					}
				}
			}
		}
	}

	/*
	 * Recursively checks inside of the <code>resource</code> for changes.
	 * @param resource	A resource to check, not <code>null</code>.
	 * @param parent	The parent directory of the files listed in list, 
	 * 					not <code>null</code>.
	 * @param list		A list of simple file names that exist in the 
	 * 					parent directory, not <code>null</code>.
	 * @param result	A result to record changes, not <code>null</code>.
	 * @return boolean	<code>true</code> if and only if the resource 
	 * 					was found in the list of resources.
	 */
	private boolean searchInResource(IResource resource, File parent, String[] list, PollingResult result) {
		String name= resource.getName();
		for (int i = 0; i < list.length; i++) {
			if (name.equals(list[i]) && isRunning()) {
				checkResource(resource, new File(parent, list[i]), result);
				return true;
			}
		}
		return false;	
	}

	/*
	 * @param name			A simple resource name, not <code>null</code>.
	 * @param resources		A list of resources, not <code>null</code>.
	 * @return boolean		<code>true</code>, if and only if one of the
	 * 						resources' simple name is equal to the given name.
	 */
	private boolean resourceExists(String name, IResource[] resources) {
		for (int i = 0; i < resources.length; i++) {
			if (name.equals(resources[i].getName())) {
				return true;
			}
		}
		return false;	
	}	

	/*
	 * Checks a file to see if it has changed or transformed.
	 * 
	 * @param resource	The resource corresponding to the file, not <code>null</code>.
	 * @param file		The file that may have changed, not <code>null</code>.
	 * @param result	A result torecord changes, not <code>null</code>.
	 */
	private void checkFile(IFile resource, File file, PollingResult result) {
		if (file.isDirectory()) {
			result.recordTransform(resource);
			return;
		}
		if (resource instanceof org.eclipse.core.internal.resources.File) {
			org.eclipse.core.internal.resources.File fileResource= (org.eclipse.core.internal.resources.File)resource;
			ResourceInfo info= fileResource.getResourceInfo(false, false);
			if (info == null) {
				result.recordDeletion(resource);
				return;
			}
			if (fileResource.isLocal(info.getFlags(), 0)) {				
				if (!fileResource.isSynchronized(IResource.DEPTH_ZERO)) {
					result.recordChange(resource);
				}
			} else {
				if (file.exists()) {
					result.recordChange(resource);
				}
			}
		}
	}

	/*
	 * Checks to see if a container exists.
	 * @param container		The container corresponding to the file, not <code>null</code>.
	 * @param file			The file that may not exist, not <code>null</code>.
	 * @param result		A result to record the deletion, not <code>null</code>.
	 */
	private void checkContainerExistence(IContainer container, File file, PollingResult result) {
		if (!file.exists()) {
			result.recordDeletion(container);
		}
	}	
}
