/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.internal;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IResource;

import com.example.autorefresh.IRefreshRequest;

/**
 * The <code>PollingResult</code> is used to collect the findings
 * during a poll.
 */
public class PollingResult implements IRefreshRequest {
	/**
	 * The map of resource to list of filenames which represent additions.
	 */
	private Map fAdditions;
	/**
	 * The set of deletions.
	 */
	private Set fDeletions;
	/**
	 * The set of changes.
	 */
	private Set fChanges;
	/**
	 * The set of transforms.
	 */
	private Set fTransforms;

	/*
	 * @see ISearchCollector#recordAddition(IResource, String)
	 */	
	public void recordAddition(IResource parent, String addition) {
		if (addition == null || parent == null)
			return;
		if (fAdditions == null) {
			fAdditions= new HashMap(1);
		}
		List additions= (List)fAdditions.get(parent);
		if (additions == null) {
			additions= new ArrayList(1);
			fAdditions.put(parent, additions);
		}
		additions.add(addition);
	}
	
	/*
	 * @see ISearchCollector#recordDeletion(IResource)
	 */	
	public void recordDeletion(IResource deletion) {
		if (deletion == null)
			return;
		if (fDeletions == null) {
			fDeletions= new HashSet(1);	
		}
		fDeletions.add(deletion);
	}
	
	/*
	 * @see ISearchCollector#recordChange(IResource)
	 */	
	public void recordChange(IResource change) {
		if (change == null)
			return;
		if (fChanges == null) {
			fChanges= new HashSet(1);
		}
		fChanges.add(change);
	}

	/*
	 * @see ISearchCollector#recordTransform(IResource)
	 */
	public void recordTransform(IResource transform) {
		if (transform == null)
			return;
		if (fTransforms == null) {
			fTransforms= new HashSet(1);
		}
		fTransforms.add(transform);
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getAdditions()
	 */
	public Map getAdditions() {
		return fAdditions;
	}
	
	/*
	 * @see com.example.autorefresh.IRefreshRequest#getDeletions()
	 */
	public IResource[] getDeletions() {
		if (fDeletions != null && !fDeletions.isEmpty()) {
			return (IResource[])fDeletions.toArray(new IResource[fDeletions.size()]);
		} 
		return null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getChanges()
	 */
	public IResource[] getChanges() {
		if (fChanges != null && !fChanges.isEmpty()) {
			return (IResource[])fChanges.toArray(new IResource[fChanges.size()]);
		} 
		return null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#getTransforms()
	 */
	public IResource[] getTransforms() {
		if (fTransforms != null && !fTransforms.isEmpty()) {
			return (IResource[])fTransforms.toArray(new IResource[fTransforms.size()]);
		} 
		return null;
	}
	
	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasAdditions()
	 */
	public boolean hasAdditions() {
		return fAdditions != null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasChanges()
	 */
	public boolean hasChanges() {
		return fChanges != null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasDeletions()
	 */
	public boolean hasDeletions() {
		return fDeletions != null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#hasTransforms()
	 */
	public boolean hasTransforms() {
		return fTransforms != null;
	}

	/*
	 * @see com.example.autorefresh.IRefreshRequest#isEmpty()
	 */
	public boolean isEmpty() {
		if (hasAdditions())
			return false;
		if (hasDeletions())
			return false;
		if (hasChanges())
			return false;
		if (hasTransforms())
			return false;
		return true;
	}

}
