/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.internal;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.internal.resources.Workspace;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.IWorkspaceDescription;
import org.eclipse.core.resources.IWorkspaceRunnable;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Preferences;
import org.eclipse.core.runtime.Preferences.PropertyChangeEvent;

import com.example.autorefresh.AutoRefreshPlugin;
import com.example.autorefresh.IRefreshRequest;

/**
 * The <code>RefreshManager</code> class refreshes resources when necessary.
 * 
 * New resources to be refreshed can be passed in through
 * a <code>IRefreshRequest</code>.  Resources will be refreshed if they
 * are added, deleted, changed or transformed.
 * <p>
 * Added resources exist in the filesystem but do not yet exist in
 * the Workbench's model.
 * <p>
 * Deleted resources exist in the Workbench's model but no longer
 * exist in the filesystem.
 * <p>
 * Changed resources have been modified in some way other than
 * by addition, deletion or transformation.
 * <p>
 * Transformed resources have transformed from file to folder or
 * from folder to file.
 */
public class RefreshManager implements Runnable {
	/*
	 * Flag to control the running state of the manager
	 */
	private boolean fRunning;
	
	/*
	 * List of refresh requests
	 */
	private List fRequests;
	
	/*
	 * Progress monitor
	 */
	private IProgressMonitor fMonitor;
	
	/*
	 * The property change listener
	 */
	private Preferences.IPropertyChangeListener fPropertyChangeListener;
	
	public RefreshManager() {
		fRequests= new ArrayList(1);
		fRunning= false;
		
		fPropertyChangeListener= new Preferences.IPropertyChangeListener() {
			/**
			 * Notifies the <code>RefreshManager</code> when the PREF_AUTO_POLL
			 * property changes.
			 *
			 * @see IPropertyChangeListener#propertyChange(PropertyChangeEvent)
			 */
			public void propertyChange(PropertyChangeEvent event) {
				if (event.getProperty().equals(AutoRefreshPlugin.PREF_AUTO_REFRESH)) {
					manageAutoRefresh();
				}
			}
		};
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		preferences.addPropertyChangeListener(fPropertyChangeListener);
		
	}

	private void manageAutoRefresh() {
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		boolean autoRefresh= preferences.getBoolean(AutoRefreshPlugin.PREF_AUTO_REFRESH);
		if (autoRefresh) {
			start();
		} else {
			stop();
		}		
	}
	
	/**
	 * Starts the refresh manager
	 */
	public void start() {
		if (!fRunning) {
			fRunning= true;
			new Thread(this, "Refresh Manager Thread").start();
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Starting Refresh Manager");
			}
		}
	}
	
	/**
	 * Shuts down the refresh manager
	 */
	public void stop() {
		if (fRunning) {
			fRunning= false;
			synchronized (this) {
				notifyAll();
			}
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Stopping Refresh Manager");
			}
		}		
	}
	
	/**
	 * 
	 */
	public void shutdown() {
		Preferences preferences= AutoRefreshPlugin.getDefault().getPluginPreferences();
		preferences.removePropertyChangeListener(fPropertyChangeListener);
		fPropertyChangeListener= null;
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Shutdown Refresh Manager");
		}
		
	}		
	
	/**
	 * Adds the refresh request to the list of refresh requests
	 * to be refreshed the next time a background refresh occurs,
	 * or an explicit refresh request is made through refresh(IProgressMonitor) 
	 * @param request a refresh request, not <code>null</code>.
	 */
	public synchronized void addRequest(IRefreshRequest request) {
		fRequests.add(request);
	}
	
	/**
	 * Requests that an immeadiate refresh of all requests occur.
	 * @param monitor for monitoring progress, or <code>null</code>.
	 */
	public void refresh(IProgressMonitor monitor) {
		fMonitor= monitor;
		if (fRunning) {
			synchronized (this) {
				notifyAll();
			}
		} else {
			refresh();
		}
	}
	
	/*
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		while(fRunning) {
			try {
				long delay= AutoRefreshPlugin.getDefault().getPluginPreferences().getLong(AutoRefreshPlugin.PREF_AUTO_REFRESH_DELAY);
				synchronized (this) {
					wait(delay);
				}
			} catch (InterruptedException e) {
				fRunning= false;
			}
			if (fRunning) {
				refresh();
			}
		}
	}	
	
	/*
	 * Iterates over the list of refresh requests refreshing.
	 * Uses the progress monitor specified in the refresh(IProgressMonitor)
	 * call, otherwise, uses a null progress monitor to report progress.
	 */
	private void refresh() {
		IProgressMonitor monitor= fMonitor;
		fMonitor= null;

		List requests= null;
		synchronized (this) {
			if (!fRequests.isEmpty()) {
				requests= fRequests;
				fRequests= new ArrayList(1);
			}
		}
		if (requests != null) {
			for (Iterator i = requests.iterator(); i.hasNext();) {
				IRefreshRequest request = (IRefreshRequest) i.next();
				refresh(request, monitor);
			}
		}
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Refreshed at " + System.currentTimeMillis());
		}
	}
		
	/**
	 * Refreshes all of the resources contained in the request.
	 * <p>
	 * <b>NOTE:</b> Turns autobuild off briefly while the resources
	 * are refreshed so that background refreshes do not cause builds.
	 * 
	 * @param request	A refresh request, can be <code>null</code>.
	 * @param monitor	A progress monitor, can be <code>null</code>.
	 */
	private void refresh(final IRefreshRequest request, IProgressMonitor monitor) {
		if (request == null || isEmpty(request)) {
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.err.println("NOT REFRESHING " + request);
			}
			return;
		}

		IWorkspaceRunnable run= new IWorkspaceRunnable() {
			public void run(IProgressMonitor monitor) throws CoreException {
				if (AutoRefreshPlugin.getDefault().isDebugging()) {
					System.out.println("Refresh starting");
				}
				long time= System.currentTimeMillis();
				((Workspace)ResourcesPlugin.getWorkspace()).getWorkManager().avoidAutoBuild();
				refresh0(request, monitor);
				time= System.currentTimeMillis() - time;
				if (AutoRefreshPlugin.getDefault().isDebugging()) {
					System.out.println("Refresh finished in: " + time + "ms");
				}
			}
		};
		
		try {
			IWorkspace workspace= ResourcesPlugin.getWorkspace();
			if (workspace != null) {
				workspace.run(run, monitor);
			} else {
				if (AutoRefreshPlugin.getDefault().isDebugging()) {
					System.out.println(this + " could not get workspace");
				}
			}
		} catch(CoreException e) {
			AutoRefreshPlugin.log(e, "Problems encountered performing refresh");
		}
	}
	
	/**
	 * Refreshes the resources found in the request.
	 */
	private void refresh0(IRefreshRequest request, IProgressMonitor monitor) {
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println("Refreshing: " + request);
		}
		if (hasDepthZeroResources(request)) {
			refreshDepthZeroResources(request, monitor);
		}		
		if (hasDepthOneResources(request)) {
			refreshDepthOneResources(request, monitor);
		}
		if (hasDepthInfiniteResources(request)) {
			refreshDepthInfiniteResources(request, monitor);
		} 
	}		

	/*
	 * @param request	The request to be polled for resources.
	 * @return boolean	<code>true</code> if and only if the request
	 * 					contains resources which need to be refreshed to
	 * 					IResource.DEPTH_ZERO.
	 */
	private boolean hasDepthZeroResources(IRefreshRequest request) {
		return request.hasChanges() || request.hasDeletions();
	}
	
	/*
	 * Answers the resources that need to be refreshed to depth
	 * IResource.DEPTH_ZERO.
	 * 
	 * @param request		The request to be polled for the resources.
	 * @return IResource[]	The resources that need refreshing, can
	 * 						be <code>null</code>.
	 * @see org.eclipse.core.resources.IResource
	 */
	private IResource[] getDepthZeroResources(IRefreshRequest request) {
		IResource[] changes= request.getChanges();
		IResource[] deletions= request.getDeletions();
		if (changes == null) {
			return deletions;
		}
		if (deletions == null) {
			return changes;
		}
		IResource[] total= new IResource[changes.length + deletions.length];
		System.arraycopy(changes, 0, total, 0, changes.length);
		System.arraycopy(deletions, 0, total, changes.length, deletions.length);
		return total;
	}
	
	/*
	 * Refreshes the resources contained in the request which need to 
	 * be refreshed to depth IResource.DEPTH_ZERO.
	 * @param request	The request to be polled for resources.
	 * @param monitor	A progress monitor, can be <code>null</code>.
	 */
	private void refreshDepthZeroResources(IRefreshRequest request, IProgressMonitor monitor) {
		refreshAll(getDepthZeroResources(request), IResource.DEPTH_ZERO, monitor);
	}	
	
	/*
	 * @param request	The request to be polled for resources.
	 * @return boolean	<code>true</code> if and only if the request
	 * 					contains resources which need to be refreshed to
	 * 					IResource.DEPTH_ONE.
	 */
	private boolean hasDepthOneResources(IRefreshRequest request) {
		return request.hasAdditions();
	}
	
	/*
	 * Answers the resources that need to be refreshed to depth
	 * IResource.DEPTH_ONE.
	 * 
	 * @param request		The request to be polled for the resources.
	 * @return IResource[]	The resources that need refreshing, can
	 * 						be <code>null</code>.
	 * @see org.eclipse.core.resources.IResource
	 */

	private Map getDepthOneResources(IRefreshRequest request) {
		return request.getAdditions();
	}
	
	/*
	 * Refreshes the resources contained in the request which need to 
	 * be refreshed to depth IResource.DEPTH_ONE.
	 * @param request	The request to be polled for resources.
	 * @param monitor	A progress monitor, can be <code>null</code>.
	 */
	private void refreshDepthOneResources(IRefreshRequest request, IProgressMonitor monitor) {
		Map depthOne= getDepthOneResources(request);
		if (depthOne == null)
			return;
		for(Iterator i= depthOne.keySet().iterator(); i.hasNext(); ) {
			IContainer container= (IContainer)i.next();

			// Optimization: only refresh resources that don't already
			// exist, this improves multiple refreshes on the same resource
			List additions= (List)depthOne.get(container);
			for(Iterator j= additions.iterator(); j.hasNext(); ) {
				String name= (String)j.next();
				if (container.findMember(name) != null)
					j.remove();
			}
			try {
				container.refreshLocal(IResource.DEPTH_ONE, monitor);
			} catch(CoreException e) {
				AutoRefreshPlugin.log(e, "Could not refresh local on " + container);
				return;
			}
			IResource[] members= null;
			try {
				members= container.members();
			} catch(CoreException e) {
				AutoRefreshPlugin.log(e, "Could not obtain the members of " + container);
				return;
			}
			if (members != null) {
				for(Iterator j= additions.iterator(); j.hasNext(); ) {
					String name= (String)j.next();
					IResource member= findMember(members, name);
					if (member instanceof IContainer) {
						try {
							member.refreshLocal(IResource.DEPTH_INFINITE, monitor);
						} catch(CoreException e) {
							AutoRefreshPlugin.log(e, "Could not refresh local on " + member);
						}
					}
				}
			}
		}
	}	

	/*
	 * @param request	The request to be polled for resources.
	 * @return boolean	<code>true</code> if and only if the request
	 * 					contains resources which need to be refreshed to
	 * 					IResource.DEPTH_INFINITE.
	 */
	private boolean hasDepthInfiniteResources(IRefreshRequest request) {
		return request.hasTransforms();
	}
	
	/*
	 * Refreshes the resources contained in the request which need to 
	 * be refreshed to depth IResource.DEPTH_INFINITE.
	 * @param request	The request to be polled for resources.
	 * @param monitor	A progress monitor, can be <code>null</code>.
	 */
	private void refreshDepthInfiniteResources(IRefreshRequest request, IProgressMonitor monitor) {
		refreshAll(getDepthInfiniteResources(request), IResource.DEPTH_INFINITE, monitor);
	}
		
	/*
	 * Answers the resources that need to be refreshed to depth
	 * IResource.DEPTH_INFINITE.
	 * 
	 * @param request		The request to be polled for the resources.
	 * @return IResource[]	The resources that need refreshing, can
	 * 						be <code>null</code>.
	 * @see org.eclipse.core.resources.IResource
	 */
	private IResource[] getDepthInfiniteResources(IRefreshRequest request) {
		return request.getTransforms();
	}
	
	/*
	 * Searches through the <code>members</code> array for a member
	 * whose simple name is <code>name</code>.
	 */
	private IResource findMember(IResource[] members, String name) {
		for (int i = 0; i < members.length; i++) {
			IResource member= members[i];
			if (name.equals(member.getName()))
				return member;
		}
		return null;
	}
	
	/*
	 * Enables or disables autobuilding.
	 */
	private void setAutoBuilding(IWorkspace workspace, boolean autoBuilding) {
		IWorkspaceDescription description= workspace.getDescription();
		description.setAutoBuilding(autoBuilding);
		try {
			workspace.setDescription(description);
		} catch(CoreException e) {
			AutoRefreshPlugin.log(e, "Could not set auto building to " + autoBuilding);
		}
	}
	
	/*
	 * Answers <code>true</code> if and only if the request has no
	 * additions, changes, deletions or transforms.
	 */
	private boolean isEmpty(IRefreshRequest request) {
		return !request.hasAdditions() && !request.hasChanges() && !request.hasDeletions() && !request.hasTransforms();
	}
	
	
	/*
	 * Refreshes all of the resources to the given depth.
	 */
	private void refreshAll(IResource[] resources, int depth, IProgressMonitor monitor) {
		if (resources == null) {
			return;
		}
		for (int i= 0, length= resources.length; i < length; i++) {
			IResource resource= resources[i];
			try {
				resource.refreshLocal(depth, monitor);
				if (AutoRefreshPlugin.getDefault().isDebugging()) {
					System.out.println(this + " refreshed " + resource);
				}
			} catch (CoreException e) {
				AutoRefreshPlugin.log(e, "Could not refresh local on " + resource);
			}
		}
	}
}
