/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.linux;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;

import com.example.autorefresh.AutoRefreshPlugin;
import com.example.autorefresh.IMonitor;

public class Monitor implements Runnable, IMonitor {

	protected IMonitorListener listener;
	protected boolean running;
	
	/* List of HandleContainers, storing references to explicitly monitored
	 * directories so we can cancel those monitors */
	protected List containers;

	class HandleContainer {
		String path;
		int handle;
		HandleContainer(String path, int handle) {
			this.path = path;
			this.handle = handle;
		}
	}

	public Monitor(IMonitorListener listener) {
		this.listener = listener;
		this.containers = new ArrayList(20);
		running = false;
	}

	/* We block on FAM events in here (not, note, polling FAM), 
	 * so this should be run in a seperate thread */
	public void run() {
		running = true;
		while (running) {
			long[] handles = getHandles();
			int nCount = handles.length;
			MonitorEvent event;
			
			/* Block on the next filesystem event from FAM */
			event = LinuxNatives.GetNextEvent();
			
			if (event == null || !running) {
				/* null is only returned when a failure to communicate with FAM occurs:
				 * in this situation the best thing to do is to drop all monitors and try again */
				running = false;
			} else {
				if (event.getEventType() == MonitorEvent.EVENT_TYPE_CHANGED) {
					listener.fileChanged(event.getFilename());
				} else if (event.getEventType() == MonitorEvent.EVENT_TYPE_CREATED) {
					File file = new File (event.getFilename());
					if (file.isDirectory() == true) {
						/* monitor the newly created path too, since monitors are recursive */
						addPath (file);
						listener.directoryCreated(event.getFilename());
					} else {
						listener.fileCreated(event.getFilename());
					}
				} else if (event.getEventType() == MonitorEvent.EVENT_TYPE_DELETED) {
					File file = new File (event.getFilename());
					if (file.isDirectory() == true) {
						/* stop monitoring this path, it doesn't exist anymore :-) */
						 removePath (file);
					}	
									
					listener.fileDeleted(event.getFilename());
				} else if (event.getEventType() == MonitorEvent.EVENT_TYPE_SHUTDOWN) {
					/* FAM has been switched off */
					running = false;	
				}
			}
		}
	}

	/**
	 * Method done.
	 * 
	 * Stop monitoring for filesystem events.
	 */
	public synchronized void done() {
		if (!running) return;
		running = false;
	}

	/**
	 * Method addPath.
	 * 
	 * Monitor path and all sub-directories for filesystem changes.
	 * 
	 * @param path path to monitor, must be a directory
	 * @return boolean true if the monitor was succesfully added
	 */
	public synchronized boolean addPath(String path) {
		File location = new File (path);
		return addPath (location);
	}

	/* Worker function that actually does the recursive monitor addition */
	protected boolean addPath(File location) {
		if (!running) return false;

		int handle;

		if (location.isDirectory()) {
			File[] files = location.listFiles();
			for (int i=0; i < files.length; i++) {
				if (files[i].isDirectory()) {
					addPath (files[i]);
				}
			}

			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println ("Added monitor for: " + location);
			}
			handle = LinuxNatives.AddDirectoryMonitor (location.getPath());
		} else {
			/* we don't handle file monitors for now */
			return false;	
		}

		if (handle == 0) {
			/* Handle of 0 means the natives failed to add a monitor for this location */
			return false;
		} else {
			/* Keep track of the handle so we can later remove it from monitoring */
			HandleContainer container = new HandleContainer(location.getPath(), handle);
			containers.add(container);
			return true;
		}
	}

	/**
	 * Method removePath.
	 * 
	 * Stop monitoring path
	 * 
	 * @param path path to be removed from monitoring
	 */
	public synchronized void removePath(String path) {
		File location = new File (path);
		removePath (location);	
	}

	protected void removePath(File location) {
		if (!running)
			return;

		if (location.isDirectory()) {
			File[] files = location.listFiles();
			for (int i=0; i < files.length; i++) {
				if (files[i].isDirectory()) {
					removePath (files[i]);
				}
			}
		}			
			
		HandleContainer container = removeHandleContainer(location.getPath());
		if (container == null)
			return;
		if (!LinuxNatives.RemoveMonitor(container.handle)) {
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Problem closing handle: " + location);
			}
		} else {
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println("Handle closed");
			}
		}
	}
	
	protected HandleContainer removeHandleContainer(String path) {
		for (int i = 0; i < containers.size(); i++) {
			HandleContainer container = (HandleContainer) containers.get(i);
			if (container.path.equals(path))
				return (HandleContainer) containers.remove(i);
		}
		return null;
	}
	
	protected synchronized HandleContainer removeHandleContainer(long handle) {
		for (int i = 0; i < containers.size(); i++) {
			HandleContainer container = (HandleContainer) containers.get(i);
			if (container.handle == handle)
				return (HandleContainer) containers.remove(i);
		}
		return null;
	}
	
	protected synchronized HandleContainer getHandleContainer(long handle) {
		for (int i = 0; i < containers.size(); i++) {
			HandleContainer container = (HandleContainer) containers.get(i);
			if (container.handle == handle)
				return container;
		}
		return null;
	}

	protected synchronized long[] getHandles() {
		long[] result = new long[containers.size()];
		for (int i = 0; i < containers.size(); i++) {
			HandleContainer container = (HandleContainer) containers.get(i);
			result[i] = container.handle;
		}
		return result;
	}
	
	public void unmonitor(IResource resource) {
		if (resource != null) {
			IPath path= resource.getLocation();
			if (path != null) {
				removePath(path.toOSString());
			} else {
				AutoRefreshPlugin.log(null, resource + " is not managed locally.");
			}
		} else {
			if (containers != null && !containers.isEmpty()) {
				HandleContainer[] containerArray= (HandleContainer[])containers.toArray(new HandleContainer[containers.size()]);
				for (int i= 0; i < containerArray.length; i++) {
					HandleContainer container= containerArray[i];
					removePath(container.path);
				}
			}
		}
	}
}