/*
 * Copyright (c) 2002 IBM Corp.  All rights reserved.
 * This file is made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 */
package com.example.autorefresh.ui.internal;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IPluginDescriptor;
import org.eclipse.core.runtime.Preferences;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.ListenerList;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.ui.IStartup;
import org.eclipse.ui.plugin.AbstractUIPlugin;

import com.example.autorefresh.AutoRefreshPlugin;

/**
 * The <code>AutorefreshUIPlugin</code> exists only to start the
 * <code>AutorefreshPlugin</code>.
 * 
 * @see org.eclipse.ui.IStartup
 */
public class AutorefreshUIPlugin extends AbstractUIPlugin implements IStartup {
	/*
	 * Singleton instance of the plugin
	 */
	private static AutorefreshUIPlugin fDefault;
	
	/*
	 * Storage for preferences, copied from AbstractUIPlugin
	 */
	private CompatibilityPreferenceStore preferenceStore;	
	
	private List fPollActionDelegates;
	
	public void registerPollActionDelegate(PollActionDelegate delegate) {
		if (fPollActionDelegates == null) {
			fPollActionDelegates= new ArrayList(1);
		}
		fPollActionDelegates.add(delegate);
	}
	
	public void setAutoPoll(boolean autoPoll) {
		if (fPollActionDelegates != null && fPollActionDelegates.size() > 0) {
			PollActionDelegate[] delegates= (PollActionDelegate[])fPollActionDelegates.toArray(new PollActionDelegate[fPollActionDelegates.size()]);
			for (int i= 0; i < delegates.length; i++) {
				PollActionDelegate delegate= delegates[i];
				delegate.setVisible(autoPoll);
			}
		}
	}

	private List fRefreshActionDelegates;
	
	public void registerRefreshActionDelegate(PollActionDelegate delegate) {
		if (fRefreshActionDelegates == null) {
			fRefreshActionDelegates= new ArrayList(1);
		}
		fRefreshActionDelegates.add(delegate);
	}
	
	public void setAutoRefresh(boolean autoRefresh) {
		if (fRefreshActionDelegates != null && fRefreshActionDelegates.size() > 0) {
			RefreshActionDelegate[] delegates= (RefreshActionDelegate[])fPollActionDelegates.toArray(new RefreshActionDelegate[fPollActionDelegates.size()]);
			for (int i= 0; i < delegates.length; i++) {
				RefreshActionDelegate delegate= delegates[i];
				delegate.setVisible(autoRefresh);
			}
		}
	}
	
	/**
	 * Default constructur for all plugins
	 */
	public AutorefreshUIPlugin(IPluginDescriptor descriptor) {
		super(descriptor);
		fDefault= this;
	}

	/**
	 * @return AutorefreshUIPlugin The singleton instance of the AutoRefreshUIPlugin.
	 */
	public static AutorefreshUIPlugin getDefault() {
		return fDefault;
	}
	
	/**
	 * @see IStartup#earlyStartup()
	 */
	public void earlyStartup() {
		AutoRefreshPlugin.getDefault().doStartup();
	}


	/**
	 * Copied from AbstractUIPlugin
	 * 
	 * @see org.eclipse.ui.plugin.AbstractUIPlugin#getPreferenceStore()
	 */
	public IPreferenceStore getPreferenceStore() {
		// Create the preference store lazily.
		if (preferenceStore == null) {
			// must assign field before calling initialize(), since
			// this method can be reentered during initialization
			preferenceStore= new CompatibilityPreferenceStore();
			// force initialization
			preferenceStore.initialize();
		}
		return preferenceStore;
	}

	/**
	 * Copied from AbstractUIPlugin
	 * 
	 * Internal implementation of a JFace preference store atop a core runtime
	 * preference store.
	 * 
	 * @since 2.0
	 */
	private class CompatibilityPreferenceStore implements IPreferenceStore {
	
		/**
		 * Flag to indicate that the listener has been added.
		 */
		private boolean listenerAdded = false;
		
		/**
		 * The underlying core runtime preference store; <code>null</code> if it
		 * has not been initialized yet.
		 */
		private Preferences prefs = null;
	
		/**
		 * Identity list of old listeners (element type: 
		 * <code>org.eclipse.jface.util.IPropertyChangeListener</code>).
		 */
		private ListenerList listeners = new ListenerList();
	
		/**
		 * Indicates whether property change events should be suppressed
		 * (used in implementation of <code>putValue</code>). Initially
		 * and usually <code>false</code>.
		 * 
		 * @see IPreferenceStore#putValue
		 */
		private boolean silentRunning = false;
	
		/**
		 * Creates a new instance for the this plug-in.
		 */
		public CompatibilityPreferenceStore() {
			// Important: do not call initialize() here
			// due to heinous reentrancy problems.
		}
		
		/**
		 * Initializes this preference store.
		 */
		void initialize() {
			// ensure initialization is only done once.
			if (this.prefs != null) {
				return;
			}
			// here's where we first ask for the plug-in's core runtime 
			// preferences;
			// note that this causes this method to be reentered
			this.prefs = AutoRefreshPlugin.getDefault().getPluginPreferences();
			// avoid adding the listener a second time when reentered
			if (!this.listenerAdded) {
				// register listener that funnels everything to firePropertyChangeEvent
				this
					.prefs
					.addPropertyChangeListener(new Preferences.IPropertyChangeListener() {
					public void propertyChange(Preferences.PropertyChangeEvent event) {
						if (!silentRunning) {
							firePropertyChangeEvent(
								event.getProperty(),
								event.getOldValue(),
								event.getNewValue());
						}
					}
				});
				this.listenerAdded = true;
			}
		}
	
		/**
		 * Returns the underlying preference store.
		 * 
		 * @return the underlying preference store
		 */
		private Preferences getPrefs() {
			if (prefs == null) {
				// although we try to ensure initialization is done eagerly,
				// this cannot be guaranteed, so ensure it is done here
				initialize();
			}
			return prefs;
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void addPropertyChangeListener(final IPropertyChangeListener listener) {
			listeners.add(listener);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void removePropertyChangeListener(IPropertyChangeListener listener) {
			listeners.remove(listener);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void firePropertyChangeEvent(
			String name,
			Object oldValue,
			Object newValue) {
	
			// efficiently handle case of 0 listeners
			if (listeners.isEmpty()) {
				// no one interested
				return;
			}
	
			// important: create intermediate array to protect against listeners 
			// being added/removed during the notification
			Object[] list = listeners.getListeners();
			PropertyChangeEvent event =
				new PropertyChangeEvent(this, name, oldValue, newValue);
			for (int i = 0; i < list.length; i++) {
				((IPropertyChangeListener) list[i]).propertyChange(event);
			}
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public boolean contains(String name) {
			return getPrefs().contains(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public boolean getBoolean(String name) {
			return getPrefs().getBoolean(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public boolean getDefaultBoolean(String name) {
			return getPrefs().getDefaultBoolean(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public double getDefaultDouble(String name) {
			return getPrefs().getDefaultDouble(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public float getDefaultFloat(String name) {
			return getPrefs().getDefaultFloat(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public int getDefaultInt(String name) {
			return getPrefs().getDefaultInt(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public long getDefaultLong(String name) {
			return getPrefs().getDefaultLong(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public String getDefaultString(String name) {
			return getPrefs().getDefaultString(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public double getDouble(String name) {
			return getPrefs().getDouble(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public float getFloat(String name) {
			return getPrefs().getFloat(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public int getInt(String name) {
			return getPrefs().getInt(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public long getLong(String name) {
			return getPrefs().getLong(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public String getString(String name) {
			return getPrefs().getString(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public boolean isDefault(String name) {
			return getPrefs().isDefault(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public boolean needsSaving() {
			return getPrefs().needsSaving();
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void putValue(String name, String value) {
			try {
				// temporarily suppress event notification while setting value
				silentRunning = true;
				getPrefs().setValue(name, value);
			} finally {
				silentRunning = false;
			}
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, double value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, float value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, int value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, long value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, String value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setDefault(String name, boolean value) {
			getPrefs().setDefault(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setToDefault(String name) {
			getPrefs().setToDefault(name);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, double value) {
			getPrefs().setValue(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, float value) {
			getPrefs().setValue(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, int value) {
			getPrefs().setValue(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, long value) {
			getPrefs().setValue(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, String value) {
			getPrefs().setValue(name, value);
		}
	
		/* (non-javadoc)
		 * Method declared on IPreferenceStore
		 */
		public void setValue(String name, boolean value) {
			getPrefs().setValue(name, value);
		}
	}	
}
