package com.example.autorefresh.win32.internal;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;

import com.example.autorefresh.AutoRefreshPlugin;
import com.example.autorefresh.IMonitor;

/**
 * A monitor that works on Win32 platforms.  Provides simple
 * notification of entire trees by reporting that the root of the
 * tree has changed to depth DEPTH_INFINITE.
 */
public class Win32Monitor implements IMonitor, Runnable {
	/**
	 * The listener that recieves callbacks and error notifications.
	 */
	protected IMonitorListener fListener;
	/**
	 * Flag to stop the monitor.
	 */
	protected boolean fIsRunning;
	/**
	 * Mapping of handles (java.lang.Long) to absolute 
	 * paths (java.lang.String).
	 */
	protected Map fHandlePathMap;
	/**
	 * Arrays of handles, split evenly when the number of
	 * handles is larger than Win32Natives.MAXIMUM_WAIT_OBJECTS
	 */
	protected long[][] fHandleArrays;
	/**
	 * Creates a new monitor.
	 * @param listener	A listener which will recieve refresh callbacks
	 * 					and error notifications, can not be <code>null</code>.
	 */
	public Win32Monitor(IMonitorListener listener) {
		fListener = listener;
		fHandlePathMap = new HashMap(1);
		fIsRunning= false;
		setHandleArrays(createHandleArrays());
	}

	/*
	 * @see com.example.autorefresh.IMonitor#unmonitor(IContainer)
	 */
	public void unmonitor(IResource resource) {
		if (resource == null) {
			// remove all containers
			synchronized (this) {
				long[][] handleArrays= getHandleArrays();
				for (int i= 0; i < handleArrays.length; i++) {
					long[] handles= handleArrays[i];
					for (int j = 0; j < handles.length; j++) {
						closeHandle(handles[j]);
					}
				}
				setHandleArrays(createHandleArrays());
			}
		} else {
			IPath path= resource.getLocation();
			if (path != null) {
				String osPath= path.toOSString();
				if (osPath != null) {
					removeHandle(getHandle(osPath));
					return;
				}
			}
		}
	}
	
	/**
	 * Adds the path to the wrappered monitor.
	 * @param path	A path to be monitored, cannot be <code>null</code>.
	 */
	public boolean addResource(IResource resource) {
		String path= resource.getLocation().toOSString();
		
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println(this + ": creating handle for " + path);
		}
		if (path == null) {
			// cannot monitor remotely managed containers
			return false;
		}

		synchronized (this) {
			if (fHandlePathMap.containsValue(path)) {
				fListener.logError(new Exception(), this + ": already monitoring " + path);
				return false;
			}
			
			long handle= Win32Natives.FindFirstChangeNotification(path, true, Win32Natives.FILE_NOTIFY_CHANGE_FILE_NAME | Win32Natives.FILE_NOTIFY_CHANGE_DIR_NAME | Win32Natives.FILE_NOTIFY_CHANGE_LAST_WRITE);
			if (handle == Win32Natives.INVALID_HANDLE_VALUE) {
				int error= Win32Natives.GetLastError();
				fListener.logError(new Exception(), this + ": could not create handle for \"" + path + "\".  Error: " + error);
				return false;
			}
			
			fHandlePathMap.put(new Long(handle), path);
			setHandleArrays(createHandleArrays());
		}
		
		if (!fIsRunning) {
			fIsRunning= true;
			new Thread(this, "Win32 Filesystem Monitor").start();
		}
		
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println(this + ": handle created for " + path);
		}
		
		return true;		
	}
	
	/**
	 * Stops the monitor
	 */
	public void stop() {
		fIsRunning= false;
	}
	
	/*
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println(this + ": started");
		}

		while (fIsRunning) {
			long[][] handleArrays= getHandleArrays();
			if (handleArrays.length == 0) {
				stop();
				break;
			}
			// If changes occur to the list of handles, 
			// ignore them until the next time through the loop.
			for (int i= 0, length= handleArrays.length; i < length && fIsRunning; i++) {
				waitFor(handleArrays[i]);
			}
		}
		if (AutoRefreshPlugin.getDefault().isDebugging()) {
			System.out.println(this + ": stopped");
		}
	}
	
	/**
	 * Performs the native call to wait for notification on one
	 * of the given handles.  
	 * @param handles an array of handles, it must contain no duplicates.
	 */
	private void waitFor(long[] handles) {
		int handleCount= handles.length;
		int index= Win32Natives.WaitForMultipleObjects(handleCount, handles, false, 300);
		if (index == Win32Natives.WAIT_TIMEOUT) {
			// nothing happened.
			return;
		}
		if (index == Win32Natives.WAIT_FAILED) {
			// we ran into a problem
			int error= Win32Natives.GetLastError();
			if (error != Win32Natives.ERROR_INVALID_HANDLE &&
				error != Win32Natives.ERROR_SUCCESS) {
				fListener.logError(new Exception(), this + ": WaitForMultipleObjects failure: " + error);
				stop();
				AutoRefreshPlugin.getDefault().monitorFailed(this, null);
			}
			return;
		}
		// a change occurred
		// WaitForMultipleObjects returns WAIT_OBJECT_0 + index
		index-= Win32Natives.WAIT_OBJECT_0;
		long handle= handles[index];
		String path;
		synchronized (this) {
			path= (String)fHandlePathMap.get(new Long(handle));
		}
		if (path != null) {
			fListener.refresh(path, IMonitorListener.DEPTH_INFINITE);
			if (!Win32Natives.FindNextChangeNotification(handle)) {
				int error= Win32Natives.GetLastError();
				if (error != Win32Natives.ERROR_INVALID_HANDLE ||
					error != Win32Natives.ERROR_SUCCESS) {
					fListener.logError(new Exception(), this + ": FindNextChangeNotification failure: " + error + " path: " + path);
				}
				removeHandle(handle);
			}
		} else {
			// worst case scenario:
			// The path couldn't be found, so the
			// AutoRefreshPlugin can't be notified that
			// this monitor noticed a change for the given handle. 
			removeHandle(handle);
		}
	}

	/**
	 * Invalidates the handle, removes it from the map of handles, and then
	 * initializes the handle arrays.
	 *
	 * @param handle			A file handle
	 * @return HandleContainer	The <code>HandleContainer</code> that was
	 * 							removed, or <code>null</code> if no container
	 * 							was found for the given file handle.
	 */
	private void removeHandle(long handle) {
		if (handle == Win32Natives.INVALID_HANDLE_VALUE) {
			return;
		}
		synchronized (this) {
			closeHandle(handle);
			setHandleArrays(createHandleArrays());
		}
	}
	
	/**
	 * Invalidates the handle, and removes it from the map
	 * of handles.
	 * @param handle a handle opened with FindFirstChangeNotification
	 */
	private void closeHandle(long handle) {
		if (handle == Win32Natives.INVALID_HANDLE_VALUE) {
			return;
		}
		boolean removed;
		synchronized (fHandlePathMap) {
			removed= fHandlePathMap.remove(new Long(handle)) != null; 
		}
		if (removed) {
			if (!Win32Natives.FindCloseChangeNotification(handle)) {
				int error= Win32Natives.GetLastError();
				if (error != Win32Natives.ERROR_INVALID_HANDLE) {
					fListener.logError(new Exception(), this + ": problem closing handle " + error);
				}
			}
			if (AutoRefreshPlugin.getDefault().isDebugging()) {
				System.out.println(this + ": handle removed " + handle);
			}	
		}		
	}
	
	/**
	 * Answers the handle for the given path.  Does not create a new handle.
	 * @param path		An absolute file path, can be <code>null</code>.
	 * @return long	The handle for the given path, or <code>WindowsNatives.INVALID_HANDLE_VALUE</code>
	 * 					if the given path is not being monitored.
	 */
	private long getHandle(String path) {
		if (path == null) {
			return Win32Natives.INVALID_HANDLE_VALUE;
		}
		synchronized (fHandlePathMap) {
			for (Iterator i= fHandlePathMap.keySet().iterator(); i.hasNext();) {
				Long handle= (Long) i.next();
				if (fHandlePathMap.get(handle).equals(path)) {
					return handle.longValue();
				}
			}
		}
		return Win32Natives.INVALID_HANDLE_VALUE;
	}
	
	/**
	 * Answers arrays of handles.  The handles are split
	 * evenly when the number of handles becomes larger
	 * than Win32Natives.MAXIMUM_WAIT_OBJECTS.
	 * @return long[][]
	 */
	private synchronized long[][] getHandleArrays() {
		return fHandleArrays;
	}
	
	private synchronized void setHandleArrays(long[][] arrays) {
		fHandleArrays= arrays;
	}	

	/**
	 * Splits the list of handles into arrays no larger than
	 * Win32Natives.MAXIMUM_WAIT_OBJECTS.  Each array contains
	 * the same number of elements, with the exception of the last
	 * one, which may contain slightly less.
	 */
	private long[][] createHandleArrays() {
		long[] handles;
		synchronized (fHandlePathMap) {
			Set keys= fHandlePathMap.keySet();
			int size= keys.size();
			if (size == 0) {
				return new long[0][0];
			} 
			handles= new long[size];
			int count= 0;
			for (Iterator i= keys.iterator(); i.hasNext();) {
				handles[count++]= ((Long) i.next()).longValue();
			}
		}
		return balencedSplit(handles, Win32Natives.MAXIMUM_WAIT_OBJECTS);
	}

	/*
	 * Splits the given array into arrays of length no greater than
	 * <code>max</code>.  The lengths of the sub arrays differ
	 * in size by no more than one element.  
	 * <p>
	 * Examples:
	 * <ul>
	 * <li>If an array of size 11 is split with a max of 4, the resulting arrays are of
	 * size 4, 4, and 3.</li>
	 * <li>If an array of size 18 is split with a max of 5, the resulting arrays are of
	 * size 5, 5, 4, and 4.</li>
	 * </ul>
	 */
	private final long[][] balencedSplit(final long[] array, final int max) {
		int elementCount= array.length;
		int subArrayCount= ((elementCount - 1) / max) + 1; // want to handle [1, max] rather than [0, max)
		int subArrayBaseLength= elementCount / subArrayCount;
		int overflow= elementCount % subArrayCount;
		long[][] result= new long[subArrayCount][];
		int count= 0;
		for(int i= 0; i < subArrayCount; i++) {
			int subArrayLength= subArrayBaseLength + (overflow-- > 0 ? 1 : 0);
			long[] subArray= new long[subArrayLength];
			for(int j= 0; j < subArrayLength; j++) {
				subArray[j]= array[count++];
			}
			result[i]= subArray;
		}
		return result;
	}
	
	/*
	 * For debugging only.
	 *
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Win32 Monitor (" + hashCode() + ")";
	}					
}